Package: nestable
Type: Package
Title: Collapsible 'HTML' Tables from Hierarchical Data
Version: 0.1.1
Authors@R: person("Derek", "Underwood", email = "dereku@gmail.com",
    role = c("aut", "cre"))
Description: Creates collapsible, expandable 'HTML' tables from hierarchical
    data. Supports data frame input with multi-level grouping, custom column
    formatters, bottom-up rollup aggregation, and CSS-variable-based theming.
    Works in 'Shiny' applications, R Markdown, 'Quarto', and the 'RStudio' Viewer.
License: GPL (>= 3)
URL: https://github.com/derekunderwood/nestable
BugReports: https://github.com/derekunderwood/nestable/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: htmltools (>= 0.5.0)
Suggests: shiny (>= 1.7.0), testthat (>= 3.0.0)
NeedsCompilation: no
Packaged: 2026-06-16 01:03:03 UTC; derek
Author: Derek Underwood [aut, cre]
Maintainer: Derek Underwood <dereku@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-16 02:20:02 UTC
Built: R 4.6.0; ; 2026-06-16 08:09:36 UTC; unix

# Generated by roxygen2: do not edit by hand

export(col_def)
export(df_to_tree)
export(fmt_currency)
export(fmt_percent)
export(nestable)
export(nestableOutput)
export(nestable_theme)
export(node)
export(renderNestable)
export(rows_to_nodes)
export(weighted_rollup)
importFrom(htmltools,HTML)
importFrom(htmltools,browsable)
importFrom(htmltools,div)
importFrom(htmltools,htmlDependency)
importFrom(htmltools,tagList)
importFrom(htmltools,tags)

# nestable.R — usage example
# Load the package in development, then run this file to render in the Viewer.
#
#   devtools::load_all()   # or library(nestable) after installing
#   source("nestable.R")

devtools::load_all(".")

# ---------------------------------------------------------------------------
# 1. Data — simulates a database result set
# ---------------------------------------------------------------------------

mag7 <- data.frame(
  name       = c("Apple",      "Microsoft", "Nvidia",         "Alphabet",
                 "Meta",        "Amazon",    "Tesla"),
  sector     = c("Technology", "Technology", "Technology",     "Comm. Services",
                 "Comm. Services", "Cons. Discretionary",     "Cons. Discretionary"),
  subsector  = c("Hardware",   "Software",  "Semiconductors", "Search & Ads",
                 "Social Media", "E-Commerce",                "EV & Auto"),
  market_cap = c(3270, 2990, 2640, 1870, 1480, 2180,  790),   # $B
  ytd_return = c(-11.8, -8.3, 22.1,  4.6, 17.2,  2.9, -33.4), # %
  stringsAsFactors = FALSE
)

# Pre-supplied aggregated returns (e.g. from a performance attribution system).
# These override rollup for the named nodes; any unlisted column still rolls up.
supplied_returns <- list(
  "Hardware"              = list(ytd_return = -10.5),
  "Software"              = list(ytd_return =  -7.8),
  "Semiconductors"        = list(ytd_return =  23.0),
  "Search & Ads"          = list(ytd_return =   5.1),
  "Social Media"          = list(ytd_return =  18.0),
  "E-Commerce"            = list(ytd_return =   3.2),
  "EV & Auto"             = list(ytd_return = -32.8),
  "Mag 7"                 = list(ytd_return =   2.9)
)

# group_col applied outermost-first: sector > subsector > stock
# total = "Mag 7" adds an optional grand-total root row; set NULL to remove it
data_root <- df_to_tree(mag7,
  name_col    = "name",
  value_cols  = c("market_cap", "ytd_return"),
  group_col   = c("sector", "subsector"),
  total       = "Mag 7",
  node_values = supplied_returns
)

# ---------------------------------------------------------------------------
# 2. Column definitions
# ---------------------------------------------------------------------------
# Simple form: named character vector (like kable's col.names).
# Headers come from the names; default format and rollup are applied.
#
#   columns <- c("Market Cap" = "market_cap", "YTD Return" = "ytd_return")
#
# Full form: list of col_def() for custom formatting, colours, and rollup.

columns <- list(
  col_def("market_cap",
    header = "Market Cap",
    format = fmt_currency("$", "B", digits = 1L),
    width  = "130px"                               # fixed width; prevents wrapping
  ),
  col_def("ytd_return",
    header = "YTD Return",
    format = fmt_percent(digits = 2L),
    color  = function(x) if (x >= 0) "#2e7d32" else "#c62828",
    rollup = weighted_rollup("market_cap"),
    width  = "110px"
  )
)

# ---------------------------------------------------------------------------
# 3. Theme
# ---------------------------------------------------------------------------

th <- nestable_theme(
  title      = "Magnificent 7",
  header_bg  = "#37474f",
  indent_px  = 20L,
  zoom = 1.25
)

# ---------------------------------------------------------------------------
# 4. Render — displays in RStudio Viewer or browser
# ---------------------------------------------------------------------------

nestable(data_root, columns, th, name_header = "TEST")

# ---------------------------------------------------------------------------
# Example 2: iris — general (non-financial) table
# ---------------------------------------------------------------------------

iris_data <- iris
iris_data$obs <- paste0("Obs.", seq_len(nrow(iris_data)))

# Each row is a leaf (observation); Species is the grouping level.
# Rollup uses "mean" so group rows show the average measurement.
iris_root <- df_to_tree(iris_data,
  name_col   = "obs",
  value_cols = c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width"),
  group_col  = "Species",
  total      = "All Species"
)

iris_cols <- list(
  col_def("Sepal.Length", rollup = "mean", width = "115px",
          format = function(x) formatC(x, digits = 2L, format = "f")),
  col_def("Sepal.Width",  rollup = "mean", width = "115px",
          format = function(x) formatC(x, digits = 2L, format = "f")),
  col_def("Petal.Length", rollup = "mean", width = "115px",
          format = function(x) formatC(x, digits = 2L, format = "f")),
  col_def("Petal.Width",  rollup = "mean", width = "115px",
          format = function(x) formatC(x, digits = 2L, format = "f"))
)

iris_theme <- nestable_theme(
  title       = "iris: measurements by species",
  table_max_w = "800px",
  header_bg   = "#4527a0",
  zoom        = 0.9          # scale the whole table down; useful for wide layouts
)

nestable(iris_root, iris_cols, iris_theme, name_header = "Observation")

# ---------------------------------------------------------------------------
# Shiny usage:
#
#   library(shiny)
#   ui     <- fluidPage(nestableOutput("tbl"))
#   server <- function(input, output) {
#     output$tbl <- renderNestable(nestable(data_root, columns, th))
#   }
#   shinyApp(ui, server)
# ---------------------------------------------------------------------------

#' Define a tree node
#'
#' @param name Display label shown in the Name column.
#' @param ... Child [node()] objects. Supplying children makes this a parent
#'   (group) row whose column values are rolled up from children unless
#'   overridden via `.values`.
#' @param .values Named list of column values. For leaf nodes supply all values
#'   here. For parent nodes any value supplied here overrides the computed
#'   rollup for that column; omitted columns are still computed from children.
#' @return A named list with elements `name`, `values`, and `children`.
#' @export
node <- function(name, ..., .values = list()) {
  list(name = name, values = as.list(.values), children = list(...))
}

#' Convert data frame rows into leaf nodes
#'
#' @param df A data frame.
#' @param name_col Column name to use as the node label.
#' @param value_cols Character vector of column names to carry as `.values`.
#' @return A list of [node()] objects.
#' @export
rows_to_nodes <- function(df, name_col, value_cols) {
  if (!is.character(name_col) || length(name_col) != 1L)
    stop("`name_col` must be a single column name string.", call. = FALSE)
  if (!name_col %in% names(df))
    stop("`name_col` \"", name_col, "\" not found in `df`.", call. = FALSE)
  missing_vals <- setdiff(value_cols, names(df))
  if (length(missing_vals))
    stop("Columns not found in `df`: ",
         paste(missing_vals, collapse = ", "), call. = FALSE)
  lapply(seq_len(nrow(df)), function(i) {
    row    <- df[i, , drop = FALSE]
    values <- as.list(row[, value_cols, drop = FALSE])
    node(as.character(row[[name_col]]), .values = values)
  })
}

#' Convert a flat data frame into a nested node tree
#'
#' @param df A data frame.
#' @param name_col Column name to use as the node label (leaf rows).
#' @param value_cols Character vector of value column names.
#' @param group_col Character vector of grouping columns, outermost first.
#'   Each element adds one nesting level. `NULL` returns a flat list of leaves.
#' @param total Optional string. When non-`NULL` a single root node with this
#'   label wraps the entire tree (grand-total row). `NULL` for no total.
#' @param node_values Optional named list of pre-supplied values for group
#'   (and total) nodes. Each name is a node label; each value is a named list
#'   of column values that should be displayed as-is rather than rolled up from
#'   children. Useful when aggregated figures (e.g. time-weighted returns) are
#'   already known and differ from a simple weighted average of the leaves.
#'
#'   Example — supply a pre-computed return for the "Technology" sector and the
#'   "Mag 7" grand total:
#'   ```r
#'   node_values = list(
#'     "Technology" = list(ytd_return = 2.5),
#'     "Mag 7"      = list(ytd_return = 4.1)
#'   )
#'   ```
#'   Any column *not* listed for a node still falls back to rollup from
#'   children.
#' @return A list of [node()] objects suitable for passing to [nestable()].
#' @export
df_to_tree <- function(df, name_col, value_cols,
                       group_col   = NULL,
                       total       = NULL,
                       node_values = list()) {
  if (!is.character(name_col) || length(name_col) != 1L)
    stop("`name_col` must be a single column name string.", call. = FALSE)
  if (!name_col %in% names(df))
    stop("`name_col` \"", name_col, "\" not found in `df`.", call. = FALSE)
  missing_vals <- setdiff(value_cols, names(df))
  if (length(missing_vals))
    stop("Columns not found in `df`: ",
         paste(missing_vals, collapse = ", "), call. = FALSE)
  if (!is.null(group_col)) {
    missing_grp <- setdiff(group_col, names(df))
    if (length(missing_grp))
      stop("group_col columns not found in `df`: ",
           paste(missing_grp, collapse = ", "), call. = FALSE)
  }
  result <- if (is.null(group_col) || length(group_col) == 0) {
    rows_to_nodes(df, name_col, value_cols)
  } else {
    g      <- group_col[1]
    rest   <- if (length(group_col) > 1) group_col[-1] else NULL
    groups <- unique(df[[g]])
    lapply(groups, function(grp) {
      sub        <- df[df[[g]] == grp, , drop = FALSE]
      grp_key    <- as.character(grp)
      supplied   <- if (!is.null(node_values[[grp_key]])) node_values[[grp_key]] else list()
      children   <- df_to_tree(sub, name_col, value_cols, rest,
                               node_values = node_values)
      do.call(node, c(list(name = grp_key, .values = supplied), children))
    })
  }
  if (!is.null(total)) {
    supplied <- if (!is.null(node_values[[total]])) node_values[[total]] else list()
    list(do.call(node, c(list(name = total, .values = supplied), result)))
  } else {
    result
  }
}

test_that("node() returns correct structure", {
  n <- node("root", .values = list(x = 1))
  expect_equal(n$name, "root")
  expect_equal(n$values$x, 1)
  expect_length(n$children, 0)
})

test_that("node() with children stores them", {
  child <- node("child", .values = list(x = 5))
  parent <- node("parent", child)
  expect_length(parent$children, 1)
  expect_equal(parent$children[[1]]$name, "child")
})

# ---------------------------------------------------------------------------
# rows_to_nodes
# ---------------------------------------------------------------------------
test_that("rows_to_nodes converts each row to a leaf node", {
  df <- data.frame(label = c("A", "B"), v = c(10, 20), stringsAsFactors = FALSE)
  nodes <- rows_to_nodes(df, "label", "v")
  expect_length(nodes, 2)
  expect_equal(nodes[[1]]$name, "A")
  expect_equal(nodes[[2]]$values$v, 20)
})

test_that("rows_to_nodes errors on missing name_col", {
  df <- data.frame(v = 1)
  expect_error(rows_to_nodes(df, "label", "v"), "name_col")
})

test_that("rows_to_nodes errors on missing value_cols", {
  df <- data.frame(label = "A")
  expect_error(rows_to_nodes(df, "label", "missing"), "missing")
})

test_that("rows_to_nodes errors when name_col is not a scalar string", {
  df <- data.frame(a = 1, b = 2)
  expect_error(rows_to_nodes(df, c("a", "b"), "a"), "name_col")
})

# ---------------------------------------------------------------------------
# df_to_tree
# ---------------------------------------------------------------------------
test_that("df_to_tree with no group_col produces flat leaf list", {
  df <- data.frame(nm = c("x", "y"), v = c(1, 2), stringsAsFactors = FALSE)
  roots <- df_to_tree(df, "nm", "v")
  expect_length(roots, 2)
  expect_equal(roots[[1]]$name, "x")
})

test_that("df_to_tree group_col creates one parent per group", {
  df <- data.frame(
    nm  = c("a", "b", "c"),
    grp = c("G1", "G1", "G2"),
    v   = c(1, 2, 3),
    stringsAsFactors = FALSE
  )
  roots <- df_to_tree(df, "nm", "v", group_col = "grp")
  expect_length(roots, 2)
  expect_equal(roots[[1]]$name, "G1")
  expect_length(roots[[1]]$children, 2)
})

test_that("df_to_tree total wraps everything in a single root", {
  df <- data.frame(nm = c("a", "b"), v = c(1, 2), stringsAsFactors = FALSE)
  roots <- df_to_tree(df, "nm", "v", total = "All")
  expect_length(roots, 1)
  expect_equal(roots[[1]]$name, "All")
})

test_that("df_to_tree errors on bad name_col", {
  df <- data.frame(nm = "a", v = 1)
  expect_error(df_to_tree(df, "nope", "v"), "name_col")
})

test_that("df_to_tree errors on bad value_cols", {
  df <- data.frame(nm = "a", v = 1)
  expect_error(df_to_tree(df, "nm", "nope"), "nope")
})

test_that("df_to_tree errors on bad group_col", {
  df <- data.frame(nm = "a", v = 1)
  expect_error(df_to_tree(df, "nm", "v", group_col = "grp"), "grp")
})

test_that("df_to_tree respects node_values override", {
  df <- data.frame(nm = c("a", "b"), grp = c("G", "G"), v = c(1, 2),
                   stringsAsFactors = FALSE)
  roots <- df_to_tree(df, "nm", "v", group_col = "grp",
                      node_values = list(G = list(v = 99)))
  # rollup would compute sum=3; override should give 99
  cols  <- list(col_def("v", rollup = "sum"))
  tree  <- nestable:::build_tree(roots, cols, prefix = "t")
  expect_equal(tree[[1]]$values$v, 99)
})

test_that("nestable() renders without error for a simple tree", {
  roots <- list(
    node("root",
         node("a", .values = list(v = 1)),
         node("b", .values = list(v = 2)))
  )
  cols <- list(col_def("v", rollup = "sum"))
  result <- nestable(roots, cols, uid = "test01")
  expect_s3_class(result, "shiny.tag.list")
})

test_that("nestable() sum-rolls up parent values", {
  roots <- list(
    node("root",
         node("a", .values = list(v = 3)),
         node("b", .values = list(v = 7)))
  )
  cols  <- list(col_def("v", rollup = "sum"))
  # build_tree is internal but accessible for testing
  tree  <- nestable:::build_tree(roots, cols, prefix = "t")
  expect_equal(tree[[1]]$values$v, 10)
})

test_that("nestable() name_col_width applies width style to header", {
  roots <- list(node("r", node("a", .values = list(v = 1))))
  cols  <- list(col_def("v"))
  result <- nestable(roots, cols, name_col_width = "200px", uid = "test02")
  html <- as.character(result)
  expect_match(html, "width:200px")
})

test_that("nestable() accepts character vector columns shorthand", {
  roots <- list(node("r",
                     node("a", .values = list(x = 1, y = 2)),
                     node("b", .values = list(x = 3, y = 4))))
  result <- nestable(roots, c("x", "y"), uid = "test03")
  expect_s3_class(result, "shiny.tag.list")
})

test_that("nestable() accepts named character vector for headers", {
  roots <- list(node("r", node("a", .values = list(x = 1))))
  result <- nestable(roots, c("X Value" = "x"), uid = "test04")
  html <- as.character(result)
  expect_match(html, "X Value")
})

test_that("nestable() default name_col derives 'Name' header", {
  roots <- list(node("r", node("a", .values = list(v = 1))))
  result <- nestable(roots, c("v"), uid = "test05")
  html <- as.character(result)
  expect_match(html, ">Name<")
})

test_that("nestable() name_col derives title-cased header", {
  roots <- list(node("r", node("a", .values = list(v = 1))))
  result <- nestable(roots, c("v"), name_col = "security_name", uid = "test06")
  html <- as.character(result)
  expect_match(html, "Security Name")
})

test_that("nestable() explicit name_header overrides name_col derivation", {
  roots <- list(node("r", node("a", .values = list(v = 1))))
  result <- nestable(roots, c("v"), name_col = "security_name",
                     name_header = "Asset", uid = "test07")
  html <- as.character(result)
  expect_match(html, "Asset")
  expect_false(grepl("Security Name", html))
})

# Internal rollup and ID-assignment engine — not exported.

rollup <- function(node, cols) {
  if (length(node$children) == 0) return(node)
  node$children <- lapply(node$children, rollup, cols = cols)
  child_values  <- lapply(node$children, `[[`, "values")
  for (col in cols) {
    existing <- node$values[[col$key]]
    if (!is.null(existing) && !is.na(existing)) next   # explicit value wins
    vals <- sapply(child_values, function(v) {
      x <- v[[col$key]]
      if (is.null(x)) NA_real_ else as.numeric(x)
    })
    node$values[[col$key]] <- col$rollup_fn(vals, child_values)
  }
  node
}

assign_ids <- function(node, prefix, env) {
  env$counter  <- env$counter + 1L
  node$id      <- paste0(prefix, env$counter)
  node$children <- lapply(node$children, assign_ids, prefix = prefix, env = env)
  node
}

build_tree <- function(roots, cols, prefix) {
  roots  <- lapply(roots, rollup, cols = cols)
  id_env <- new.env(parent = emptyenv())
  id_env$counter <- 0L
  lapply(roots, assign_ids, prefix = prefix, env = id_env)
}

#' Create a nestable theme
#'
#' Every argument maps to a CSS custom property (`--ntbl-*`) set inline on
#' the widget's wrapper `<div>`, so multiple instances with different themes
#' can coexist on the same page without conflict.
#'
#' @param title Character. Title shown above the table. Default `""` (no title).
#' @param font_family CSS font-family string.
#' @param font_size CSS font-size string. Default `"14px"`.
#' @param table_bg Table background colour. Default `"#ffffff"`.
#' @param table_shadow CSS box-shadow for the table.
#' @param table_radius CSS border-radius. Default `"6px"`.
#' @param table_max_w CSS max-width. Default `"680px"`.
#' @param header_bg Header row background. Default `"#37474f"`.
#' @param header_color Header row text colour. Default `"#ffffff"`.
#' @param row_border Row separator colour. Default `"#eceff1"`.
#' @param row_hover_bg Row hover background. Default `"#f9fbe7"`.
#' @param parent_weight CSS font-weight for parent rows. Default `"600"`.
#' @param toggle_color Colour of the expand/collapse arrow. Default `"#546e7a"`.
#' @param indent_px Integer pixels of indentation per nesting level. Default `20L`.
#' @param zoom CSS zoom level applied to the entire widget. Accepts any valid
#'   CSS `zoom` value: a number (`1.25`), a percentage (`"125%"`), or `"normal"`
#'   (default). Useful for global size/scale adjustments without touching
#'   individual font-size or dimension settings.
#' @return A named list of theme values.
#' @export
nestable_theme <- function(
  title         = "",
  font_family   = '-apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, sans-serif',
  font_size     = "14px",
  table_bg      = "#ffffff",
  table_shadow  = "0 1px 4px rgba(0,0,0,.12)",
  table_radius  = "6px",
  table_max_w   = "680px",
  header_bg     = "#37474f",
  header_color  = "#ffffff",
  row_border    = "#eceff1",
  row_hover_bg  = "#f9fbe7",
  parent_weight = "600",
  toggle_color  = "#546e7a",
  indent_px     = 20L,
  zoom          = "normal"
) {
  list(
    title         = title,
    font_family   = font_family,
    font_size     = font_size,
    table_bg      = table_bg,
    table_shadow  = table_shadow,
    table_radius  = table_radius,
    table_max_w   = table_max_w,
    header_bg     = header_bg,
    header_color  = header_color,
    row_border    = row_border,
    row_hover_bg  = row_hover_bg,
    parent_weight = parent_weight,
    toggle_color  = toggle_color,
    indent_px     = as.integer(indent_px),
    zoom          = as.character(zoom)
  )
}

# Internal: turn an underscore_key into a Title Case label.
pretty_header <- function(key) {
  words <- strsplit(key, "[_. ]+")[[1]]
  paste(paste0(toupper(substr(words, 1, 1)), substr(words, 2, nchar(words))),
        collapse = " ")
}

# Internal: resolve "sum" / "mean" shortcut strings to rollup functions.
resolve_rollup <- function(rollup) {
  if (is.function(rollup)) return(rollup)
  switch(rollup,
    sum  = function(vals, child_values) sum(vals,  na.rm = TRUE),
    mean = function(vals, child_values) mean(vals, na.rm = TRUE),
    stop("Unknown rollup shortcut '", rollup,
         "'. Use \"sum\", \"mean\", or a function.", call. = FALSE)
  )
}

#' Define a table column
#'
#' @param key Character. The value key — the column name in your data frame (or
#'   the name used in each node's `.values` list).
#' @param header Character. Column header text. Defaults to a title-cased
#'   version of `key` (e.g. `"market_cap"` → `"Market Cap"`).
#' @param format Function `function(x) -> character` for display formatting.
#'   Defaults to [base::format()].
#' @param color Function `function(x) -> CSS color string`, or `NULL`.
#'   Default `NULL`.
#' @param rollup How parent rows are aggregated. Either a shortcut string
#'   (`"sum"` or `"mean"`) or a function `function(vals, child_values) ->
#'   scalar`. `vals` is a numeric vector of children's values for this column;
#'   `child_values` is the full list of each child's value lists (useful for
#'   weighted aggregation via [weighted_rollup()]). Defaults to `"sum"`.
#' @param width CSS width string (e.g. `"120px"`, `"10%"`) applied to the
#'   column header and every data cell. `NULL` (default) leaves width unset.
#' @return A named list describing the column.
#' @export
col_def <- function(key,
                    header = NULL,
                    format = function(x) base::format(x),
                    color  = NULL,
                    rollup = "sum",
                    width  = NULL) {
  if (is.null(header)) header <- pretty_header(key)
  list(header    = header,
       key       = key,
       format_fn = format,
       color_fn  = color,
       rollup_fn = resolve_rollup(rollup),
       width     = width)
}

#' Currency format function factory
#'
#' Returns a formatting function for use as `format_fn` in [col_def()].
#'
#' @param prefix Character prepended before the number. Default `"$"`.
#' @param suffix Character appended after the number. Default `""`.
#' @param digits Integer decimal places. Default `2L`.
#' @param big_mark Thousands separator. Default `","`.
#' @return A function `function(x) -> character`.
#' @export
fmt_currency <- function(prefix = "$", suffix = "", digits = 2L, big_mark = ",") {
  function(x) {
    paste0(prefix,
           formatC(x, format = "f", digits = digits, big.mark = big_mark),
           suffix)
  }
}

#' Percentage format function factory
#'
#' Returns a formatting function for use as `format_fn` in [col_def()].
#'
#' @param digits Integer decimal places. Default `2L`.
#' @param plus Logical. Prefix non-negative values with `"+"`. Default `TRUE`.
#' @return A function `function(x) -> character`.
#' @export
fmt_percent <- function(digits = 2L, plus = TRUE) {
  function(x) {
    sign <- if (plus && x >= 0) "+" else ""
    paste0(sign, formatC(x, format = "f", digits = digits), "%")
  }
}

#' Weighted-average rollup function factory
#'
#' Returns a rollup function for use as `rollup_fn` in [col_def()]. Computes
#' the weighted average of `vals` using another key's values as weights.
#'
#' @param weight_key Character. The value key to use as weights (e.g.
#'   `"market_cap"`). Each child's value for this key is used as its weight.
#' @return A function `function(vals, child_values) -> numeric`.
#' @export
weighted_rollup <- function(weight_key) {
  function(vals, child_values) {
    weights <- sapply(child_values, function(v) {
      w <- v[[weight_key]]
      if (is.null(w) || is.na(w)) 0 else as.numeric(w)
    })
    total_w <- sum(weights, na.rm = TRUE)
    if (total_w == 0) NA_real_ else sum(vals * weights, na.rm = TRUE) / total_w
  }
}

test_that("col_def defaults are correct", {
  cd <- col_def("market_cap")
  expect_equal(cd$key, "market_cap")
  expect_equal(cd$header, "Market Cap")
  expect_null(cd$width)
  expect_null(cd$color_fn)
  expect_equal(cd$format_fn(1234), format(1234))
})

test_that("col_def respects explicit header", {
  cd <- col_def("v", header = "Value")
  expect_equal(cd$header, "Value")
})

test_that("col_def width is stored", {
  cd <- col_def("v", width = "150px")
  expect_equal(cd$width, "150px")
})

test_that("col_def rollup sum works", {
  cd <- col_def("v", rollup = "sum")
  expect_equal(cd$rollup_fn(c(1, 2, 3), list()), 6)
})

test_that("col_def rollup mean works", {
  cd <- col_def("v", rollup = "mean")
  expect_equal(cd$rollup_fn(c(2, 4, 6), list()), 4)
})

test_that("col_def rejects unknown rollup string", {
  expect_error(col_def("v", rollup = "median"), "Unknown rollup")
})

test_that("fmt_currency formats correctly", {
  fn <- fmt_currency("$", "B", digits = 1L)
  expect_equal(fn(1.5), "$1.5B")
})

test_that("fmt_percent prefixes positive with +", {
  fn <- fmt_percent(digits = 2L)
  expect_match(fn(3.14), "^\\+3\\.14%$")
})

test_that("fmt_percent prefixes negative with -", {
  fn <- fmt_percent(digits = 2L)
  expect_match(fn(-1.5), "^-1\\.50%$")
})

test_that("weighted_rollup computes weighted average", {
  fn <- weighted_rollup("w")
  child_values <- list(list(v = 10, w = 2), list(v = 20, w = 3))
  vals <- c(10, 20)
  result <- fn(vals, child_values)
  expect_equal(result, (10 * 2 + 20 * 3) / 5)
})

test_that("weighted_rollup returns NA when total weight is zero", {
  fn <- weighted_rollup("w")
  child_values <- list(list(v = 1, w = 0), list(v = 2, w = 0))
  expect_true(is.na(fn(c(1, 2), child_values)))
})

# Internal: accept a character vector as a shorthand column spec.
#   c("market_cap", "ytd_return")          → auto-derived headers
#   c("Market Cap" = "market_cap", ...)    → explicit headers
# A list of col_def() objects is passed through unchanged.
normalise_columns <- function(columns) {
  if (is.character(columns)) {
    nms <- names(columns)
    return(lapply(seq_along(columns), function(i) {
      key <- columns[[i]]
      hdr <- if (!is.null(nms) && nzchar(nms[[i]])) nms[[i]] else NULL
      col_def(key, header = hdr)
    }))
  }
  columns
}

# Random UID safe for HTML id attributes (must not start with a digit).
new_widget_uid <- function() {
  paste0("w", paste0(sample(c(letters, 0:9), 7L, replace = TRUE), collapse = ""))
}

nestable_dependency <- function() {
  ver <- tryCatch(utils::packageVersion("nestable"), error = function(e) "0.1.0")
  # system.file() works when installed; fall back to inst/www/ in development
  src <- system.file("www", package = "nestable")
  if (!nzchar(src)) src <- file.path(getwd(), "inst", "www")
  htmlDependency(
    name       = "nestable",
    version    = ver,
    src        = src,
    script     = "nestable.js",
    stylesheet = "nestable.css"
  )
}

# Internal: walk tree and return a list of htmltools tr tags.
render_rows <- function(nodes, cols, depth, parent_id, indent_px, name_col_width) {
  rows <- list()

  for (nd in nodes) {
    has_children <- length(nd$children) > 0
    padding_left <- depth * indent_px + 8L

    toggle <- if (has_children) {
      tags$button(
        class           = "ntbl-toggle",
        `data-target`   = nd$id,
        `aria-expanded` = "false",
        HTML("&#9654;")
      )
    } else {
      tags$span(class = "ntbl-toggle-spacer")
    }

    name_td_style <- paste0("padding-left:", padding_left, "px",
                            if (!is.null(name_col_width))
                              paste0(";width:", name_col_width,
                                     ";white-space:nowrap") else "")
    name_td <- tags$td(
      style = name_td_style,
      toggle,
      if (has_children) tags$strong(nd$name) else nd$name
    )

    value_tds <- lapply(cols, function(col) {
      val     <- nd$values[[col$key]]
      display <- if (is.null(val) || is.na(val)) HTML("&mdash;") else col$format_fn(val)
      parts   <- c(
        if (!is.null(col$width))    paste0("width:",     col$width)    else NULL,
        if (!is.null(col$width))    "white-space:nowrap"               else NULL,
        if (!is.null(col$color_fn) && !is.null(val) && !is.na(val))
          paste0("color:", col$color_fn(val))                          else NULL
      )
      style <- if (length(parts)) paste(parts, collapse = ";") else NULL
      tags$td(class = "ntbl-num", style = style, display)
    })

    row_attrs <- list(
      id    = nd$id,
      class = if (has_children) "ntbl-row ntbl-parent" else "ntbl-row ntbl-leaf"
    )
    if (!is.null(parent_id)) {
      row_attrs[["data-parent"]] <- parent_id
      row_attrs[["style"]]       <- "display:none"
    }

    rows <- c(rows, list(do.call(tags$tr, c(row_attrs, list(name_td), value_tds))))

    if (has_children) {
      rows <- c(rows, render_rows(nd$children, cols,
                                  depth          = depth + 1L,
                                  parent_id      = nd$id,
                                  indent_px      = indent_px,
                                  name_col_width = name_col_width))
    }
  }

  rows
}

#' Create a nestable collapsible HTML table
#'
#' @param data_root A list of top-level [node()] objects. Build with [node()],
#'   [rows_to_nodes()], or [df_to_tree()].
#' @param columns Column specification. Three forms are accepted:
#'   \itemize{
#'     \item A character vector of key names:
#'           `c("market_cap", "ytd_return")` — headers are auto-derived
#'           from the key (e.g. `"market_cap"` → `"Market Cap"`).
#'     \item A *named* character vector:
#'           `c("Market Cap" = "market_cap", "YTD Return" = "ytd_return")` —
#'           explicit headers, default formatting and rollup.
#'     \item A list of [col_def()] objects for full control over formatting,
#'           colours, and rollup behaviour.
#'   }
#' @param name_col Character. The node label key — the `name_col` used when
#'   building the tree with [df_to_tree()], or `"name"` when constructing nodes
#'   manually. Used to auto-derive `name_header` via title-casing when
#'   `name_header` is `NULL`. Default `"name"`.
#' @param name_header Character. Header label for the first (name/label) column.
#'   `NULL` (default) derives the label from `name_col` (e.g. `"security_name"`
#'   → `"Security Name"`).
#' @param name_col_width CSS width string (e.g. `"200px"`, `"30%"`) applied to
#'   the name column header and every name cell. `NULL` (default) leaves the
#'   width unset, allowing the browser to size the column automatically.
#' @param theme A theme list from [nestable_theme()].
#' @param uid Character. Widget UID prefix for HTML element `id` attributes.
#'   Defaults to a random string so multiple tables on one page never clash.
#'   Override only when reproducible IDs are needed (e.g. tests).
#' @return An [htmltools::browsable()] `tagList`. Renders inline in R Markdown,
#'   Quarto, and the RStudio Viewer; use inside [shiny::renderUI()] or
#'   [renderNestable()] in Shiny apps.
#' @importFrom htmltools tags div HTML browsable tagList htmlDependency
#' @export
nestable <- function(data_root,
                     columns,
                     theme          = nestable_theme(),
                     name_col       = "name",
                     name_header    = NULL,
                     name_col_width = NULL,
                     uid            = new_widget_uid()) {

  if (is.null(name_header)) name_header <- pretty_header(name_col)
  columns <- normalise_columns(columns)
  tree <- build_tree(data_root, columns, prefix = paste0(uid, "-"))

  zoom_css <- if (!is.null(theme$zoom) && theme$zoom != "normal")
    sprintf("; zoom: %s", theme$zoom) else ""

  css_vars <- paste0(paste(
    sprintf("--ntbl-font-family: %s",   theme$font_family),
    sprintf("--ntbl-font-size: %s",     theme$font_size),
    sprintf("--ntbl-table-bg: %s",      theme$table_bg),
    sprintf("--ntbl-table-shadow: %s",  theme$table_shadow),
    sprintf("--ntbl-table-radius: %s",  theme$table_radius),
    sprintf("--ntbl-table-max-w: %s",   theme$table_max_w),
    sprintf("--ntbl-header-bg: %s",     theme$header_bg),
    sprintf("--ntbl-header-color: %s",  theme$header_color),
    sprintf("--ntbl-row-border: %s",    theme$row_border),
    sprintf("--ntbl-row-hover-bg: %s",  theme$row_hover_bg),
    sprintf("--ntbl-parent-weight: %s", theme$parent_weight),
    sprintf("--ntbl-toggle-color: %s",  theme$toggle_color),
    sep = "; "
  ), zoom_css)

  name_th_style <- if (!is.null(name_col_width))
    paste0("width:", name_col_width, ";white-space:nowrap") else NULL

  header_cells <- lapply(columns, function(col) {
    hdr_style <- if (!is.null(col$width))
      paste0("width:", col$width, ";white-space:nowrap") else NULL
    tags$th(class = "ntbl-num", style = hdr_style, col$header)
  })

  body_rows <- render_rows(tree, columns,
                           depth          = 0L,
                           parent_id      = NULL,
                           indent_px      = theme$indent_px,
                           name_col_width = name_col_width)

  widget <- div(
    id    = uid,
    class = "ntbl-wrap",
    style = css_vars,
    if (nchar(theme$title) > 0) tags$h2(class = "ntbl-title", theme$title),
    tags$table(
      class = "ntbl",
      tags$thead(tags$tr(tags$th(style = name_th_style, name_header), header_cells)),
      tags$tbody(body_rows)
    )
  )

  browsable(tagList(nestable_dependency(), widget))
}

#' Shiny UI output for a nestable table
#'
#' Use with [renderNestable()] in the server. These are thin wrappers over
#' [shiny::uiOutput()] and [shiny::renderUI()] — no `htmlwidgets` dependency
#' is required.
#'
#' @param outputId The output variable name.
#' @param ... Additional arguments passed to [shiny::uiOutput()].
#' @return A Shiny UI element.
#' @export
nestableOutput <- function(outputId, ...) {
  if (!requireNamespace("shiny", quietly = TRUE))
    stop("Package 'shiny' is required. Install with install.packages('shiny').",
         call. = FALSE)
  shiny::uiOutput(outputId, ...)
}

#' Shiny render function for a nestable table
#'
#' @param expr An expression returning a [nestable()] widget.
#' @param env The environment in which to evaluate `expr`.
#' @param quoted Logical. Is `expr` already quoted? Default `FALSE`.
#' @return A Shiny render function.
#' @rdname nestableOutput
#' @export
renderNestable <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!requireNamespace("shiny", quietly = TRUE))
    stop("Package 'shiny' is required. Install with install.packages('shiny').",
         call. = FALSE)
  if (!quoted) expr <- substitute(expr)
  shiny::renderUI(expr, env = env, quoted = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nestable.R
\name{nestable}
\alias{nestable}
\title{Create a nestable collapsible HTML table}
\usage{
nestable(
  data_root,
  columns,
  theme = nestable_theme(),
  name_col = "name",
  name_header = NULL,
  name_col_width = NULL,
  uid = new_widget_uid()
)
}
\arguments{
\item{data_root}{A list of top-level \code{\link[=node]{node()}} objects. Build with \code{\link[=node]{node()}},
\code{\link[=rows_to_nodes]{rows_to_nodes()}}, or \code{\link[=df_to_tree]{df_to_tree()}}.}

\item{columns}{Column specification. Three forms are accepted:
\itemize{
\item A character vector of key names:
\code{c("market_cap", "ytd_return")} — headers are auto-derived
from the key (e.g. \code{"market_cap"} → \code{"Market Cap"}).
\item A \emph{named} character vector:
\code{c("Market Cap" = "market_cap", "YTD Return" = "ytd_return")} —
explicit headers, default formatting and rollup.
\item A list of \code{\link[=col_def]{col_def()}} objects for full control over formatting,
colours, and rollup behaviour.
}}

\item{theme}{A theme list from \code{\link[=nestable_theme]{nestable_theme()}}.}

\item{name_col}{Character. The node label key — the \code{name_col} used when
building the tree with \code{\link[=df_to_tree]{df_to_tree()}}, or \code{"name"} when constructing nodes
manually. Used to auto-derive \code{name_header} via title-casing when
\code{name_header} is \code{NULL}. Default \code{"name"}.}

\item{name_header}{Character. Header label for the first (name/label) column.
\code{NULL} (default) derives the label from \code{name_col} (e.g. \code{"security_name"}
→ \code{"Security Name"}).}

\item{name_col_width}{CSS width string (e.g. \code{"200px"}, \code{"30\%"}) applied to
the name column header and every name cell. \code{NULL} (default) leaves the
width unset, allowing the browser to size the column automatically.}

\item{uid}{Character. Widget UID prefix for HTML element \code{id} attributes.
Defaults to a random string so multiple tables on one page never clash.
Override only when reproducible IDs are needed (e.g. tests).}
}
\value{
An \code{\link[htmltools:browsable]{htmltools::browsable()}} \code{tagList}. Renders inline in R Markdown,
Quarto, and the RStudio Viewer; use inside \code{\link[shiny:renderUI]{shiny::renderUI()}} or
\code{\link[=renderNestable]{renderNestable()}} in Shiny apps.
}
\description{
Create a nestable collapsible HTML table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/node.R
\name{rows_to_nodes}
\alias{rows_to_nodes}
\title{Convert data frame rows into leaf nodes}
\usage{
rows_to_nodes(df, name_col, value_cols)
}
\arguments{
\item{df}{A data frame.}

\item{name_col}{Column name to use as the node label.}

\item{value_cols}{Character vector of column names to carry as \code{.values}.}
}
\value{
A list of \code{\link[=node]{node()}} objects.
}
\description{
Convert data frame rows into leaf nodes
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nestable.R
\name{nestableOutput}
\alias{nestableOutput}
\alias{renderNestable}
\title{Shiny UI output for a nestable table}
\usage{
nestableOutput(outputId, ...)

renderNestable(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{The output variable name.}

\item{...}{Additional arguments passed to \code{\link[shiny:htmlOutput]{shiny::uiOutput()}}.}

\item{expr}{An expression returning a \code{\link[=nestable]{nestable()}} widget.}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Logical. Is \code{expr} already quoted? Default \code{FALSE}.}
}
\value{
A Shiny UI element.

A Shiny render function.
}
\description{
Use with \code{\link[=renderNestable]{renderNestable()}} in the server. These are thin wrappers over
\code{\link[shiny:htmlOutput]{shiny::uiOutput()}} and \code{\link[shiny:renderUI]{shiny::renderUI()}} — no \code{htmlwidgets} dependency
is required.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/node.R
\name{node}
\alias{node}
\title{Define a tree node}
\usage{
node(name, ..., .values = list())
}
\arguments{
\item{name}{Display label shown in the Name column.}

\item{...}{Child \code{\link[=node]{node()}} objects. Supplying children makes this a parent
(group) row whose column values are rolled up from children unless
overridden via \code{.values}.}

\item{.values}{Named list of column values. For leaf nodes supply all values
here. For parent nodes any value supplied here overrides the computed
rollup for that column; omitted columns are still computed from children.}
}
\value{
A named list with elements \code{name}, \code{values}, and \code{children}.
}
\description{
Define a tree node
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/column.R
\name{fmt_percent}
\alias{fmt_percent}
\title{Percentage format function factory}
\usage{
fmt_percent(digits = 2L, plus = TRUE)
}
\arguments{
\item{digits}{Integer decimal places. Default \code{2L}.}

\item{plus}{Logical. Prefix non-negative values with \code{"+"}. Default \code{TRUE}.}
}
\value{
A function \verb{function(x) -> character}.
}
\description{
Returns a formatting function for use as \code{format_fn} in \code{\link[=col_def]{col_def()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/column.R
\name{weighted_rollup}
\alias{weighted_rollup}
\title{Weighted-average rollup function factory}
\usage{
weighted_rollup(weight_key)
}
\arguments{
\item{weight_key}{Character. The value key to use as weights (e.g.
\code{"market_cap"}). Each child's value for this key is used as its weight.}
}
\value{
A function \verb{function(vals, child_values) -> numeric}.
}
\description{
Returns a rollup function for use as \code{rollup_fn} in \code{\link[=col_def]{col_def()}}. Computes
the weighted average of \code{vals} using another key's values as weights.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/column.R
\name{fmt_currency}
\alias{fmt_currency}
\title{Currency format function factory}
\usage{
fmt_currency(prefix = "$", suffix = "", digits = 2L, big_mark = ",")
}
\arguments{
\item{prefix}{Character prepended before the number. Default \code{"$"}.}

\item{suffix}{Character appended after the number. Default \code{""}.}

\item{digits}{Integer decimal places. Default \code{2L}.}

\item{big_mark}{Thousands separator. Default \code{","}.}
}
\value{
A function \verb{function(x) -> character}.
}
\description{
Returns a formatting function for use as \code{format_fn} in \code{\link[=col_def]{col_def()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/node.R
\name{df_to_tree}
\alias{df_to_tree}
\title{Convert a flat data frame into a nested node tree}
\usage{
df_to_tree(
  df,
  name_col,
  value_cols,
  group_col = NULL,
  total = NULL,
  node_values = list()
)
}
\arguments{
\item{df}{A data frame.}

\item{name_col}{Column name to use as the node label (leaf rows).}

\item{value_cols}{Character vector of value column names.}

\item{group_col}{Character vector of grouping columns, outermost first.
Each element adds one nesting level. \code{NULL} returns a flat list of leaves.}

\item{total}{Optional string. When non-\code{NULL} a single root node with this
label wraps the entire tree (grand-total row). \code{NULL} for no total.}

\item{node_values}{Optional named list of pre-supplied values for group
(and total) nodes. Each name is a node label; each value is a named list
of column values that should be displayed as-is rather than rolled up from
children. Useful when aggregated figures (e.g. time-weighted returns) are
already known and differ from a simple weighted average of the leaves.

Example — supply a pre-computed return for the "Technology" sector and the
"Mag 7" grand total:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{node_values = list(
  "Technology" = list(ytd_return = 2.5),
  "Mag 7"      = list(ytd_return = 4.1)
)
}\if{html}{\out{</div>}}

Any column \emph{not} listed for a node still falls back to rollup from
children.}
}
\value{
A list of \code{\link[=node]{node()}} objects suitable for passing to \code{\link[=nestable]{nestable()}}.
}
\description{
Convert a flat data frame into a nested node tree
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme.R
\name{nestable_theme}
\alias{nestable_theme}
\title{Create a nestable theme}
\usage{
nestable_theme(
  title = "",
  font_family = "-apple-system, BlinkMacSystemFont, \\"Segoe UI\\", Roboto, sans-serif",
  font_size = "14px",
  table_bg = "#ffffff",
  table_shadow = "0 1px 4px rgba(0,0,0,.12)",
  table_radius = "6px",
  table_max_w = "680px",
  header_bg = "#37474f",
  header_color = "#ffffff",
  row_border = "#eceff1",
  row_hover_bg = "#f9fbe7",
  parent_weight = "600",
  toggle_color = "#546e7a",
  indent_px = 20L,
  zoom = "normal"
)
}
\arguments{
\item{title}{Character. Title shown above the table. Default \code{""} (no title).}

\item{font_family}{CSS font-family string.}

\item{font_size}{CSS font-size string. Default \code{"14px"}.}

\item{table_bg}{Table background colour. Default \code{"#ffffff"}.}

\item{table_shadow}{CSS box-shadow for the table.}

\item{table_radius}{CSS border-radius. Default \code{"6px"}.}

\item{table_max_w}{CSS max-width. Default \code{"680px"}.}

\item{header_bg}{Header row background. Default \code{"#37474f"}.}

\item{header_color}{Header row text colour. Default \code{"#ffffff"}.}

\item{row_border}{Row separator colour. Default \code{"#eceff1"}.}

\item{row_hover_bg}{Row hover background. Default \code{"#f9fbe7"}.}

\item{parent_weight}{CSS font-weight for parent rows. Default \code{"600"}.}

\item{toggle_color}{Colour of the expand/collapse arrow. Default \code{"#546e7a"}.}

\item{indent_px}{Integer pixels of indentation per nesting level. Default \code{20L}.}

\item{zoom}{CSS zoom level applied to the entire widget. Accepts any valid
CSS \code{zoom} value: a number (\code{1.25}), a percentage (\code{"125\%"}), or \code{"normal"}
(default). Useful for global size/scale adjustments without touching
individual font-size or dimension settings.}
}
\value{
A named list of theme values.
}
\description{
Every argument maps to a CSS custom property (\verb{--ntbl-*}) set inline on
the widget's wrapper \verb{<div>}, so multiple instances with different themes
can coexist on the same page without conflict.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/column.R
\name{col_def}
\alias{col_def}
\title{Define a table column}
\usage{
col_def(
  key,
  header = NULL,
  format = function(x) base::format(x),
  color = NULL,
  rollup = "sum",
  width = NULL
)
}
\arguments{
\item{key}{Character. The value key — the column name in your data frame (or
the name used in each node's \code{.values} list).}

\item{header}{Character. Column header text. Defaults to a title-cased
version of \code{key} (e.g. \code{"market_cap"} → \code{"Market Cap"}).}

\item{format}{Function \verb{function(x) -> character} for display formatting.
Defaults to \code{\link[base:format]{base::format()}}.}

\item{color}{Function \verb{function(x) -> CSS color string}, or \code{NULL}.
Default \code{NULL}.}

\item{rollup}{How parent rows are aggregated. Either a shortcut string
(\code{"sum"} or \code{"mean"}) or a function \verb{function(vals, child_values) -> scalar}. \code{vals} is a numeric vector of children's values for this column;
\code{child_values} is the full list of each child's value lists (useful for
weighted aggregation via \code{\link[=weighted_rollup]{weighted_rollup()}}). Defaults to \code{"sum"}.}

\item{width}{CSS width string (e.g. \code{"120px"}, \code{"10\%"}) applied to the
column header and every data cell. \code{NULL} (default) leaves width unset.}
}
\value{
A named list describing the column.
}
\description{
Define a table column
}
